/*

 pc64.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>

#ifndef WIN32
#include <sys/io.h>
#else
#include <windows.h>
#endif

#include <unistd.h>

#include "pc64.h"
#include "crc8.h"

static unsigned short int IOPORT, STATPORT;
static int pc64s;
extern volatile char tout;

#ifdef WIN32
  #ifndef WIN32_IO

/* prototype (function typedef) for DLL function Inp32: */

typedef short _stdcall (*inpfuncPtr)(short portaddr);
typedef void _stdcall (*oupfuncPtr)(short portaddr, short datum);

HINSTANCE hLib;
inpfuncPtr inp32;
oupfuncPtr oup32;

static void outb(unsigned char value, int port)
{
    (oup32)(port,value);
}

static unsigned char inb(int port)
{
    return (inp32)(port);
}
  #else

static __inline__ void
outb(unsigned char _data, unsigned short _port)
{
    __asm__ __volatile__ ("outb %1, %0"
  :
  : "dN" (_port),
    "a" (_data));
}

static __inline__ unsigned char
inb(unsigned short _port)
{
  unsigned char rv;
  __asm__ __volatile__ ("inb %1, %0"
	  : "=a" (rv)
	  : "dN" (_port));
  return rv;
}

  #endif // !WIN32_IO
#else // WIN32
static int parportfd = -1;

#include <linux/ppdev.h>
#include <linux/parport.h>
#include <sys/ioctl.h>
#include <fcntl.h>
static int getbparport(void);
static int sendbparport(unsigned char b);
#endif // WIN32

static int vtime=1;

static int getb(void);
static int sendb(unsigned char b);

static char *drivername(void) {
    return "PC64";
}

static int initialize(char *dev, int port) {

    if (port & 0x10000) vtime=0; //hog
    pc64s = port & 0x40000;
    IOPORT = port & 0xffff;
    STATPORT = IOPORT+1;
  #ifdef WIN32
   #ifndef WIN32_IO
    /* Load the library */
    hLib = LoadLibrary("inpout32.dll");
    if (hLib == NULL) {fprintf(stderr,"LoadLibrary Failed.\n");exit(1);}

    /* get the address of the function */

    inp32 = (inpfuncPtr) GetProcAddress(hLib, "Inp32");
    if (inp32 == NULL) {fprintf(stderr,"GetProcAddress for Inp32 Failed.\n");exit(1);}

    oup32 = (oupfuncPtr) GetProcAddress(hLib, "Out32");
    if (oup32 == NULL) {fprintf(stderr,"GetProcAddress for Oup32 Failed.\n");exit(1);}
   #endif // !WIN32_IO
  #else
    if (port & 0x20000) {
        parportfd = open(dev, O_RDWR);
        if (parportfd == -1) {
            fprintf(stderr,"Opening parport device\n");
            return -ENODEV;
        }
        if (ioctl(parportfd, PPEXCL) || ioctl(parportfd, PPCLAIM)) {
            close(parportfd);
            fprintf(stderr,"Cannot access parport device\n");
            return -ENODEV;
        }
        {
            int mode=IEEE1284_MODE_COMPAT;
            if (ioctl(parportfd, PPSETMODE, &mode)) {
                close(parportfd);
                fprintf(stderr,"Cannot set byte mode on parport device\n");
                return -ENODEV;
            }

            mode=0;
            if (ioctl(parportfd, PPDATADIR, &mode)) {
                close(parportfd);
                fprintf(stderr,"Cannot set direction on parport device\n");
                return -ENODEV;
            }
        }
        pc64_io_ops.getb=getbparport;
        pc64_io_ops.sendb=sendbparport;
    } else {
        parportfd = -1;
        if (ioperm(IOPORT,2,1)) {
            fprintf(stderr,"Cannot allocate IO ports\n");
            return -ENODEV;
        }
        pc64_io_ops.getb=getb;
        pc64_io_ops.sendb=sendb;
    }
  #endif // WIN32 (LPT port)    
    return 0;
}

#ifndef WIN32
static int getbparport(void) {
    unsigned char b,a;
    unsigned char c;
    alarm(1);
    
    for (;;) {
        ioctl(parportfd, PPRSTATUS, &b);
        if (b & PARPORT_STATUS_BUSY) break;
        if (tout) return -EIO;
    }
    if (!pc64s) {
        c=0x80;ioctl(parportfd, PPWDATA, &c);
        ioctl(parportfd, PPRSTATUS, &b);
        c=0x00;ioctl(parportfd, PPWDATA, &c);
    } else {
        ioctl(parportfd, PPRSTATUS, &b);
        c=0x08;ioctl(parportfd, PPWDATA, &c);
    }
    b=(b & 0x78) >> 3;
    for (;;) {
        ioctl(parportfd, PPRSTATUS, &a);
        if (!(a & PARPORT_STATUS_BUSY)) break;
        if (tout) return -EIO;
    }
    if (!pc64s) {
        c=0x80;ioctl(parportfd, PPWDATA, &c);
    }
    ioctl(parportfd, PPRSTATUS, &a);
    c=0x00;ioctl(parportfd, PPWDATA, &c);
    b|=(a & 0x78) << 1;
    crc_add(b);
    return b;
}

static int sendbparport(unsigned char b) {
    unsigned char a;
    crc_add(b);
    alarm(1);
    if (!pc64s) {
        a=(b & 0x0f) | 0x80;
    } else {
        a=(b & 0x07);
    }
    for (;;) {
	unsigned char c;
	ioctl(parportfd, PPRSTATUS, &c);
	if (c & PARPORT_STATUS_BUSY) break;
	if (tout) return -EIO;
    }
    ioctl(parportfd, PPWDATA, &a);
    if (!pc64s) {
        a&=0x0f;ioctl(parportfd, PPWDATA, &a);
        b=(b >> 4) | 0x80;
    } else {
        a|=8;ioctl(parportfd, PPWDATA, &a);
        a=(((b >> 3) & 3) | (b & 0x04)) ^ 0x0d;
    }
    for (;;) {
	unsigned char c;
	ioctl(parportfd, PPRSTATUS, &c);
	if (!(c & PARPORT_STATUS_BUSY)) break;
	if (tout) return -EIO;
    }
    if (!pc64s) {
        ioctl(parportfd, PPWDATA, &b);
        b&=0x0f;ioctl(parportfd, PPWDATA, &b);
    } else {
        ioctl(parportfd, PPWDATA, &a);
        a=(b >> 5) ^ 3 ^ ((b >> 2) & 1);
        for (;;) {
            unsigned char c;
            ioctl(parportfd, PPRSTATUS, &c);
            if (!(c & PARPORT_STATUS_ERROR)) break;
            if (tout) return -EIO;
        }
        ioctl(parportfd, PPWDATA, &a);
    }
    return 0;
}
#endif

static int getb(void) {
    unsigned char b,a;
    alarm(1);
    
    while (inb(STATPORT)<0x80) if (tout) return -EIO;
    if (!pc64s) {
        outb(0x80,IOPORT);
        b=inb(STATPORT);
        outb(0x00,IOPORT);
        b=(b & 0x78) >> 3;
        while (inb(STATPORT)>=0x80) if (tout) return -EIO;
        outb(0x80,IOPORT);
    } else {
        b=inb(STATPORT);
        outb(0x08,IOPORT);
        b=(b & 0x78) >> 3;
        while (inb(STATPORT)>=0x80) if (tout) return -EIO;
    }
    a=inb(STATPORT);
    outb(0x00,IOPORT);
    b|=(a & 0x78) << 1;
    crc_add(b);
    return b;
}

static int sendb(unsigned char b) {
    unsigned char a;
    crc_add(b);
    alarm(1);
    if (!pc64s) {
        a=(b & 0x0f) | 0x80;
        while (inb(STATPORT)<0x80) if (tout) return -EIO;
        outb(a,IOPORT);
        outb(a & 0x0f,IOPORT);
        b=(b >> 4) | 0x80;
        while (inb(STATPORT)>=0x80) if (tout) return -EIO;
        outb(b,IOPORT);
        outb(b & 0x0f,IOPORT);
    } else {
        a=(b & 0x07);
        while (inb(STATPORT)<0x80) if (tout) return -EIO;
        outb(a,IOPORT);
        outb(a | 8,IOPORT);
        a=(((b >> 3) & 3) | (b & 0x04)) ^ 0x0d;
        while (inb(STATPORT)>=0x80) if (tout) return -EIO;
        outb(a,IOPORT);
        a=(b >> 5) ^ 3 ^ ((b >> 2) & 1);
        while (inb(STATPORT) & 0x08) if (tout) return -EIO;
        outb(a,IOPORT);
    }
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=pc64_io_ops.getb();
        if (j<0) return j;
        data[i]=j;
    }
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=pc64_io_ops.sendb(data[i]);
        if (j) return j;
    }
    return 0;
}

static void shutdown(void) {
#ifndef WIN32
    if (parportfd >= 0) {
        ioctl(parportfd, PPRELEASE);
        close(parportfd);
        parportfd = -1;
    }
#endif
    return;
}

static int flush(void) {
    alarm(0);
    return 0;
}

static void clean(void) {
    return;
}

static int wait(void) {
    int dyntime=1;
    int i;
start:
    alarm(0);
#ifndef WIN32
    if (parportfd >= 0) {
        unsigned char b;
        ioctl(parportfd, PPRDATA, &b);
        b&=~8;ioctl(parportfd, PPWDATA, &b);
        for (;;) {
            unsigned char b;
            ioctl(parportfd, PPRSTATUS, &b);
            if (b & PARPORT_STATUS_BUSY) {
                ioctl(parportfd, PPRSTATUS, &b);
                if (b & 0x78) break;
            }
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
        }
    } else
#endif
    {
        outb(inb(IOPORT) & (~8),IOPORT);
        do {
            while (inb(STATPORT)<0x80) if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
        } while ((inb(STATPORT) & 0x78)==0);
    }
    tout=0;
    i = pc64_io_ops.getb();
    if (tout) goto start;
    alarm(0);
    return (unsigned char)i;
}

struct io_ops pc64_io_ops = {
    .drivername = drivername,
    .initialize = initialize,
    .getb       = getb,
    .getbytes   = getbytes,
    .sendb      = sendb,
    .sendbytes  = sendbytes,
    .shutdown   = shutdown,
    .flush      = flush,
    .wait       = wait,
    .clean      = clean,
};
